# pi-topSPEAKER Software Brief [![Build Status](https://jenkins.pi-top.com/buildStatus/icon?job=pitop/pt-speaker-deb/master)](https://jenkins.pi-top.com/job/pitop/job/pt-speaker-deb/job/master/)

## Overview

### Package Info
**Dependencies:** i2c-tools, wiringPi
### Files
#### pt-speaker
`pt-speaker` is a BASH script, which is designed to run on startup. It checks if I2C is enabled, and then enters a permanent loop which probes the state of any pi-topSPEAKERs detected on the I2C lines, and - on detection - will ensure that HDMI is set as output, and execute pt-speaker-controller to handle setting up the newly connected speaker. It also checks for changes to the system volume level, and updates the speaker volume accordingly.

#### pt-speaker-controller
`pt-speaker-controller` is a C++ application which executes the code described in `playback.configure` (I2C Script to Setup pi-topSPEAKER in Playback Mode), to instantiate the newly connected speaker with the necessary configuration to allow playback.

#### pt-speaker-vol
`pt-speaker-vol` is a C++ application which is designed to update the speakers with the current system-level volume, including muting.

# pi-topSPEAKER Hardware Brief

## Overview

Each pi-topSPEAKER is manufactured identically, but it is a modular device and up to 3 can be connected in a chain. pi-topSPEAKER needs to be initialised each time after power up in order to function, the default register settings are not compatible with the incoming I2S audio data (and even the speaker driver is shut down by default). This is common with most software controlled devices that are highly customisable, and although a supervisor MCU could be placed on the board to initialise default settings automatically, this is undesirable due to system cost constraints - therefore the Raspberry Pi takes this role.

To avoid overlapping of I2C addresses there is a 3-position hardware switch that the user can set to select L, M or R - left channel, mono mix or right channel respectively. This switch changes the primary I2C address of the pi-topSPEAKER, this means there is a direct mapping between the I2C address seen using I2C Detect and the channel that the user wishes to be output through that particular speaker.

This hardware switch controls the I2C address on an I2C MUX, which sits between the Raspberry Pi I2C bus and the DAC+AMP of the pi-topSPEAKER. The TCA9543A is a 1:2 multiplexer, allowing I2C to be switched between two channels: 0 and 1. However, in this application it is simply being used as a 1:1 I2C address translator (actual I2C address translators are more expensive), so only channel 0 on the multiplexer is used (channel 1 is not connected to anything).

To communicate with a pi-topSPEAKER, it has to be done through the I2C MUX. You can think of the I2C MUX as a proxy, first you have to open up the proxy by writing which channel you wish to be opened up (always channel 0). This then connects the pi-topSPEAKER brain directly to the I2C bus, and communication can be carried out as normal. The proxy must be closed before trying to communicate with other pi-topSPEAKERs.

Finally, the audio output on the Raspberry Pi needs to be set to HDMI. The default parameters calculated for the internal clock dividers within TAS2505, namely DOSR, NDAC and MDAC, are compatible with 44.1, 48 and 96 kHz audio sampling rate frequencies without any changes. Anything outside of this range, e.g. 192 kHz, will require a change to these. Device performance is not optimal with these calculated values (for example, NDAC is preferred as high as possible), but the flexibility in audio output frequencies is desirable without the need to detect the audio frequency every time it is played to confirm it will work. The following reasoning confirms this:

- If detection, and then subsequent changes of divider parameters, took too long the user would would hear glitchy audio output whilst it happens.
- There are numerous ways that audio can be output and capturing this reliably every time in the OS before it is passed to lower level audio driver hardware would be troublesome.

## Block Diagram

The Raspberry Pi is set to output audio data through HDMI. This is converted by STDP2600 on the Hub into SPDIF data (BCM format). The SPDIF receiver then converts this to 24-bit I2S format, whose output clock frequencies - SCK0, BCK0 and LRCK0 - are integer multiples of the incoming audio sampling rate (44.1 kHz, 48 kHz etc).

![Diagram 1](documentation/image01-23.png)


DOUT contains both left and right audio data (providing both are being output from the Raspberry Pi). So even though pi-topSPEAKER is a mono device, the data is available and the TAS2505 allows mixing/multiplexing of this data prior to the mono DAC - this is all programmable through registers.

The mixed/multiplexed digital audio data is then input to TAS2505's mono DAC Signal Processor. The signal processing blocks are programmable, and can be tuned for high noise rejection and low group delay. There is also various signal-processing effects such as audio effects and frequency shaping that can be programmed. The AINL and AINR pins on the TAS2505 are connected directly to two pins on the PCB Rail,  they are analogue audio inputs that can be channeled directly to TAS2505's amplifier circuitry - they are shutdown by default and do not interfere with the RPi GPIOs that they are connected to on the PCB Rail (testing required). These will be used in the future by connecting a pi-topSPEAKER to the left and right side of pi-topPROTO's PCB rail connectors, so users can make a portable stereo speaker system from analogue audio data that is more understandable from a learning perspective (ham radio, synth, bluetooth MP3 player etc). The only caveat being that we will have to supply a programmed MCU that initialises the pi-topSPEAKER's settings.

The TCA9543A acts as an I2C address translator, so the user can select which channel they want a particular speaker to output using the SP3T switch - the address detected from the Raspberry Pi will map directly to the desired speaker channel. This I2C MUX will act as a proxy for communication between the Raspberry Pi and TAS2505, this is used to initialise pi-topSPEAKER with the default pi-topOS settings, but also used to adjust volume and apply audio effects dynamically in real-time.

Finally, the audio data is amplified through a 2W D-class audio amplifier and output to the 2W (3W peak), 4 Ω speaker.

## I2C Mapping

![Diagram 2](documentation/image03-26.png)

The A1 and A0 pins of the TCA9543A are set high and low using the hardware switch at the front of pi-topSPEAKER - the mapping between addresses detected and speaker channel can be seen above. The register map of TCA9543 is very simple, and only 1 bit - B0 - needs to be adjusted. When set high it connects that particular pi-topSPEAKER to the I2C bus, and when low disconnects it. B0's default value on power up is low, i.e. the only addresses visible from I2C Detect after boot up are that of of the I2C MUXs. Once B0 is set high on TCA9543A, the TAS2505 device will be visible on I2C Detect\*, and the address of this device is 0011 000 - communication can now carry out as if it was connected directly to the I2C bus.

Once finished programming one pi-topSPEAKER, B0 must be set low by writing to the I2C MUX before attempting to communicate with any other device - otherwise there will be address overlap and they will interfere with each other.

*\* Page 0 / Register 34 states by default the device will not respond to an I2C general call (address 0x00), if this is what I2C detect uses then TAS2505 will not be visible until this register is written to 1 - so no need to panic if it's not visible, just try writing to the known address.*

## I2C Software Flow

On boot up, I2C Detect should be ran so that the available speakers can be determined. B0 is then set high for one speaker, programmed, and then B0 is set low again before moving onto the next speaker.

![Diagram 2](documentation/image05-29.png)
The same software flow should be followed for adjusting parameters in the device, such as volume control. The device has a soft-stepping function for volume changes, which will be adjusted to max so that the volume changes over a long enough period to make it sound like all speakers are increasing in volume at the same time - rather than each one hard stepping as its volume is changed.

One issue that might arise is when pi-topSPEAKER's are added when the devices have already been initialised. Rather than running I2C Detect repeatedly over a specified time interval, the best option would be to run I2C Detect before any parameter adjustments are made (volume being the most common). Then if a new speaker is detected, the initialisation can be done first before changing the parameters. The user is then instructed that new devices will be automatically detected when they change the volume, or alternatively they should be able to run a "detect" command through pi-topOS dashboard.

The speaker settings should be stored in a file within pi-topOS, and whenever any register changes are made, it should first be updated in the file and then written to the device. This way, the default pi-topSPEAKER settings which are initialised to the device on boot up are always the last that the user had when they last shutdown. The standard unaltered settings file should also be stored in case the user has playback issues, then in pi-topOS dashboard there will be a "restore to default" button.

In a scenario where I2C writes are failing, one reason could be that the user could have multiple pi-topSPEAKER's set up with the same address and there is interference. A notification within the dashboard should inform them to check this, as this will be the most common mistake people will make.

## Memory Space

The memory address space of TAS2505 is set up in pages, with each page containing a number of registers. So when wanting to write to (or read from) a register within a certain page, you have to first write to the device telling it which page you want to access. Register 0x00 on every page is set up as a page select, so no matter which page you are currently in, writing to register 0x00 with the data as the next page you want to access. For example, for moving to page 0x01 you would write to register 0x00, and the data would be 0x01. The page selected is persistent, so it only needs to be done once prior to writing to registers within that page.

## Volume Control

Raspbian's volume control feature should be overtaken by pi-topOS (pt-speaker program) since much more accurate volume can be achieved using pi-topSPEAKER - testing using the standard volume control has shown large changes in volume, and below about 70% it is barely audible. Some testing is required to see which HDMI output volume results in the best performance, i.e. is it better to set the HDMI output at minimum volume, and amplify it, or is it better to set it at maximum volume and amplify/deamplify.

There are two volume controls within the DAC+AMP:

1. DAC digital volume control which implements programmable gain, adjustable from -64.5 dB to 24 dB in 0.5dB steps
2. Speaker Driver Analog output volume control:
  1. One register adjustable from 0 dB to -72.3 dB in (irregular) 0.5 dB steps.
  2. One register adjustable from mute, then from 6 dB to 24 dB in 6 dB steps

Again, testing is required to see which combination gives the best audio performance. However, the analogue output volume control is suggested to be used for output pop-noise reduction, so it may be that this will be set low and ramped up to a fixed value on audio output detection, and all adjustments will be done digitally.





## **Miscellaneous Registers that might be useful**

### Page 0 / Register 38: DAC Flag Register 2

When volume is changed, a soft stepping algorithm reduces the volume in 0.125 dB steps per sample point (or two sample points if programmed to do so) until the desired volume is reached. When this register is equal to 1 then the soft-stepping is complete and the actual volume is equal to the volume programmed to control register. This is useful if the sound is to be muted, or output drivers to be powered down, because you can be sure that the sound has reached minimum volume before you cut it - avoiding a harsh stop in sound.

### Page 0 / Register 52: GPIO/DOUT Control Register

General purpose output pin that can be programmed high or low over I2C. This will be used if the AINL and AINR pins are not 3.3V tolerant and get damaged by the Raspberry Pi setting BCM16 or BCM26 high (the pins they are connected to), by using a SPDT switch to break them from the circuit.

### Page 0 / Register 53: DOUT Function Control Register

Another General purpose output is available here, suggestions welcome.

### Page 0 / Register 55: MISO Function Control Register

Another General purpose output is available here, suggestions welcome.
